import mincepy
from mincepy.testing import *


def test_ref_list(historian: mincepy.Historian):
    """Test references list"""
    list1 = mincepy.builtins.RefList()
    list2 = mincepy.builtins.RefList()
    car = Car()
    list1.append(car)
    list2.append(car)

    # Reference condition satisfied
    assert list1[0] is list2[0]

    # Now delete everything, reload, and make sure the condition is still satisfied
    list1_id, list2_id = historian.save(list1, list2)
    del list1, list2, car

    list1_loaded = historian.load(list1_id)
    list2_loaded = historian.load(list2_id)
    assert len(list1_loaded) == 1
    assert len(list2_loaded) == 1
    assert isinstance(list1_loaded[0], Car)
    assert list1_loaded[0] is list2_loaded[0]


def test_ref_dict(historian: mincepy.Historian):
    """Test that the ref dict correctly stores entries as references"""
    dict1 = mincepy.builtins.RefDict()
    dict2 = mincepy.builtins.RefDict()
    car = Car()
    dict1['car'] = car
    dict2['car'] = car

    # Reference condition satisfied
    assert dict1['car'] is dict2['car']

    # Now delete everything, reload, and make sure the condition is still satisfied
    dict1_id, dict2_id = historian.save(dict1, dict2)
    del dict1, dict2, car

    dict1_loaded = historian.load(dict1_id)
    dict2_loaded = historian.load(dict2_id)
    assert len(dict1_loaded) == 1
    assert len(dict2_loaded) == 1
    assert isinstance(dict1_loaded['car'], Car)
    assert dict1_loaded['car'] is dict2_loaded['car']



from collections import namedtuple
from functools import partial
import logging
import typing

from bidict import bidict
import PySide2
from PySide2 import QtCore, QtGui
from PySide2.QtCore import QObject, Signal, Qt, QModelIndex

import mincepy
from . import common

__all__ = 'DbModel', 'SnapshotRecord'

logger = logging.getLogger(__name__)

DATA_RECORD_MAPPING = bidict({"[{}]".format(entry): entry for entry in mincepy.DataRecord._fields})

UNSET = ''  # The value printed for records that don't have a particular attribute

TOOLTIPS = {
    DATA_RECORD_MAPPING.inverse[mincepy.TYPE_ID]: 'Object type',
    DATA_RECORD_MAPPING.inverse[mincepy.CREATION_TIME]: 'Creation time',
    DATA_RECORD_MAPPING.inverse[mincepy.SNAPSHOT_TIME]: 'Last modification time',
    DATA_RECORD_MAPPING.inverse[mincepy.VERSION]: 'Version',
}

SnapshotRecord = namedtuple("SnapshotRecord", 'snapshot record')


def pretty_type_string(obj_type: typing.Type):
    """Given an type will return a simple type string"""
    type_str = str(obj_type)
    if type_str.startswith('<class '):
        return type_str[8:-2]
    return type_str


class DbModel(QObject):
    # Signals
    historian_changed = Signal(mincepy.Historian)

    def __init__(self):
        super().__init__()
        self._historian = None

    @property
    def historian(self) -> mincepy.Historian:
        return self._historian

    @historian.setter
    def historian(self, historian):
        self._historian = historian
        self.historian_changed.emit(self._historian)


class DataRecordQueryModel(QtCore.QAbstractTableModel):
    # Signals
    type_restriction_changed = Signal(object)
    sort_changed = Signal(dict)

    def __init__(self, db_model: DbModel, executor=common.default_executor, parent=None):
        super().__init__(parent)
        self._db_model = db_model
        self._query = {}
        self._results = None
        self._sort = None
        self._type_restriction = None
        self._column_names = mincepy.DataRecord._fields
        self._update_future = None

        self._executor = executor
        self._new_results.connect(self._inject_results)

        # If the historian changes then we get invalidated
        self._db_model.historian_changed.connect(lambda hist: self._invalidate_results())

    @property
    def db_model(self):
        return self._db_model

    @property
    def column_names(self):
        return self._column_names

    def get_query(self):
        return self._query

    def set_query(self, query: dict):
        """Set the query that will be passed to historian.find()"""
        self._query = query
        self._invalidate_results()

    def set_type_restriction(self, type_id):
        if type_id == self._type_restriction:
            return

        self._type_restriction = type_id
        self._invalidate_results()
        self.type_restriction_changed.emit(self._type_restriction)

    def get_type_restriction(self):
        return self._type_restriction

    def set_sort(self, sort):
        if sort == self._sort:
            return

        self._sort = sort
        self._invalidate_results()
        self.sort_changed.emit(self._sort)

    def get_sort(self):
        return self._sort

    def get_records(self):
        return self._results

    def refresh(self):
        self._invalidate_results()

    def rowCount(self, _parent: QtCore.QModelIndex = QModelIndex()) -> int:
        if self._results is None:
            return 0

        return len(self._results)

    def columnCount(self, _parent: QtCore.QModelIndex = QModelIndex()) -> int:
        return len(self.column_names)

    def headerData(self, section: int, orientation: PySide2.QtCore.Qt.Orientation, role: int = ...) -> typing.Any:
        if role != Qt.DisplayRole:
            return None

        if orientation == QtCore.Qt.Orientation.Horizontal:
            return self.column_names[section]

        return str(self._results[section])

    def data(self, index: PySide2.QtCore.QModelIndex, role: int = ...) -> typing.Any:
        if role == Qt.DisplayRole:
            value = getattr(self._results[index.row()], self.column_names[index.column()])
            return str(value)

        return None

    def _update_results(self):
        if self._query is None or self._db_model.historian is None:
            self._results = []
        else:
            query = self._query.copy()
            if self._type_restriction is not None:
                query['obj_type'] = self._type_restriction
            if self._sort is not None:
                query['sort'] = self._sort

            self._results = []
            self._update_future = self._executor(partial(self._perform_query, query), msg="Querying...", blocking=False)

    def _invalidate_results(self):
        self.beginResetModel()
        self._results = None
        self.endResetModel()
        self._update_results()

    def _perform_query(self, query, batch_size=4):
        logging.debug("Starting query: %s", query)

        total = 0
        batch = []
        for result in self._db_model.historian.find(**query, as_objects=False):
            batch.append(result)
            if len(batch) == batch_size:
                self._new_results.emit(batch)
                batch = []
            total += 1

        if batch:
            # Emit the last batch
            self._new_results.emit(batch)

        logging.debug('Finished query, got %s results', total)

    _new_results = Signal(list)

    @QtCore.Slot(list)
    def _inject_results(self, batch: list):
        """As a query is executed batches of results as emitted and passed to this callback for insertion"""
        first = len(self._results)
        last = first + len(batch) - 1
        self.beginInsertRows(QModelIndex(), first, last)
        self._results.extend(batch)
        self.endInsertRows()


class EntriesTable(QtCore.QAbstractTableModel):
    ARCHIVE_COLUMNS = (mincepy.OBJ_ID, mincepy.TYPE_ID, mincepy.CREATION_TIME, mincepy.SNAPSHOT_TIME, mincepy.VERSION,
                       mincepy.STATE)
    DEFAULT_COLUMNS = tuple(DATA_RECORD_MAPPING.inverse[label] for label in ARCHIVE_COLUMNS)

    def __init__(self, query_model: DataRecordQueryModel, parent=None):
        super().__init__(parent)
        self._query_model = query_model
        self._query_model.modelReset.connect(self._invalidate)
        self._query_model.rowsInserted.connect(self._query_rows_inserted)

        self._columns = list(self.DEFAULT_COLUMNS)
        self._show_objects = True

        self._snapshots_cache = {}

    def get_records(self):
        return self._query_model.get_records()

    @property
    def query_model(self):
        return self._query_model

    def get_record(self, row) -> typing.Optional[mincepy.DataRecord]:
        if row < 0 or row >= self.rowCount():
            return None

        return self.get_records()[row]

    def get_snapshot(self, row: int):
        ref = self.get_record(row).get_reference()
        if ref not in self._snapshots_cache:
            historian = self._query_model.db_model.historian
            try:
                self._snapshots_cache[ref] = historian.load_snapshot(ref)
            except TypeError:
                self._snapshots_cache[ref] = None

        return self._snapshots_cache[ref]

    def get_show_as_objects(self):
        return self._show_objects

    def set_show_as_objects(self, as_objects):
        if self._show_objects != as_objects:
            self._show_objects = as_objects
            # Remove the old columns, if there are any to be removed
            if len(self._columns) > len(self.DEFAULT_COLUMNS):
                first = len(self.DEFAULT_COLUMNS)
                last = len(self._columns) - 1

                self.beginRemoveColumns(QModelIndex(), first, last)
                self._reset_columns()
                self.endRemoveColumns()

            # Now add the new ones
            columns = set()
            for row in range(self.rowCount()):
                columns.update(self._get_columns_for(row))

            if columns:
                # Convert to list and sort alphabetically
                new_columns = list(columns)
                new_columns.sort()
                self._insert_columns(new_columns)

    def refresh(self):
        # As the query to refresh from the database
        self._query_model.refresh()

    def rowCount(self, _parent: PySide2.QtCore.QModelIndex = ...) -> int:
        return self._query_model.rowCount()

    def columnCount(self, _parent: PySide2.QtCore.QModelIndex = ...) -> int:
        return len(self._columns)

    def headerData(self, section: int, orientation: PySide2.QtCore.Qt.Orientation, role: int = ...) -> typing.Any:
        if role != Qt.DisplayRole:
            return None

        if orientation == QtCore.Qt.Orientation.Horizontal:
            if section >= len(self._columns):
                return None
            return self._columns[section]

        if orientation == QtCore.Qt.Orientation.Vertical:
            return str(section)

        return None

    def data(self, index: PySide2.QtCore.QModelIndex, role: int = ...) -> typing.Any:
        column_name = self._columns[index.column()]
        if role == Qt.DisplayRole:
            value = self._get_value_string(index.row(), index.column())
            return value
        if role == Qt.FontRole:
            if column_name in DATA_RECORD_MAPPING:
                font = QtGui.QFont()
                font.setItalic(True)
                return font
        if role == Qt.ToolTipRole:
            try:
                return TOOLTIPS[column_name]
            except KeyError:
                pass

        return None

    def sort(self, column: int, order: PySide2.QtCore.Qt.SortOrder = ...):
        column_name = self._columns[column]
        try:
            sort_criterion = DATA_RECORD_MAPPING[column_name]
        except KeyError:
            if self._show_objects:
                # We can't deal with sorting objects at the moment
                return
            sort_criterion = "state.{}".format(column_name)

        sort_dict = {sort_criterion: mincepy.ASCENDING if order == Qt.AscendingOrder else mincepy.DESCENDING}
        self._query_model.set_sort(sort_dict)

    def _invalidate(self):
        self.beginResetModel()
        self._snapshots_cache = {}
        self._reset_columns()
        self.endResetModel()

    def _get_snapshot_state(self, record):
        if self._show_objects:
            historian = self._query_model.db_model.historian
            try:
                return historian.load_snapshot(record.get_reference())
            except TypeError:
                pass  # Fall back to displaying the state

        return record.state

    def _get_columns_for(self, row: int) -> tuple:
        if self.get_show_as_objects():
            obj = self.get_snapshot(row)
            try:
                state = vars(obj)
            except TypeError:
                state = {}
        else:
            state = self.get_record(row).state

        if isinstance(state, dict):
            return tuple(state.keys())

        return ()

    def _get_value_string(self, row: int, column: int) -> str:
        column_name = self._columns[column]
        record = self.get_record(row)

        if column < len(self.DEFAULT_COLUMNS):
            original_name = DATA_RECORD_MAPPING[column_name]
            record_value = record._asdict()[original_name]

            # Special case to show type ids as the class name
            if original_name == mincepy.TYPE_ID:
                try:
                    historian = self._query_model.db_model.historian
                    return pretty_type_string(historian.get_obj_type(record_value))
                except TypeError:
                    pass
            return str(record_value)

        # The column is a custom attribute of the item
        if self.get_show_as_objects():
            snapshot = self.get_snapshot(row)
            try:
                state = vars(snapshot)
            except TypeError:
                state = {}
        else:
            state = record.state

        # The state in the record needn't be a dict so we check
        if isinstance(state, dict):
            return str(state.get(column_name, UNSET))

        return UNSET

    def _query_rows_inserted(self, _parent: QModelIndex, first: int, last: int):
        """Called when there are new entries inserted into the entries table"""

        self.beginInsertRows(QModelIndex(), first, last)

        columns = set()  # Keep track of all the columns in this batch
        for row in range(first, last + 1):
            columns.update(self._get_columns_for(row))

        self.endInsertRows()

        # Check if the columns need updating
        columns -= set(self._columns)
        if columns:
            # There are new columns to insert
            cols = list(columns)
            cols.sort()
            self._insert_columns(cols)

    def _insert_columns(self, new_columns: typing.Sequence):
        """Add new columns to our existing ones"""
        first_col = len(self._columns)
        last_col = first_col + len(new_columns) - 1
        self.beginInsertColumns(QModelIndex(), first_col, last_col)
        self._columns.extend(new_columns)
        self.endInsertColumns()

    def _reset_columns(self):
        """Reset the columns back to the default internally.  This should only be done between either
        an model invalidation or appropriate removeColumns call
        """
        self._columns = list(self.DEFAULT_COLUMNS)



from concurrent.futures import Future

import mincepy


def default_executor(func, msg=None, blocking=False):
    future = Future()
    try:
        future.set_result(func())
    except Exception as exc:
        future.set_exception(exc)
    return future


def default_create_historian(uri) -> mincepy.Historian:
    historian = mincepy.create_historian(uri)
    mincepy.set_historian(historian)
    return historian



from .archive import *
from .base_savable import *
from .archive_factory import *
from .comparators import *
from .depositors import *
from .exceptions import *
from .helpers import *
from .historian import *
from .history import *
from .process import *
from .records import *
from .refs import *
from .types import *
from .version import *
from . import analysis
from . import builtins
from . import common_helpers
from . import mongo
from . import testing

_ADDITIONAL = ('analysis', 'mongo', 'buitins', 'common_helpers', 'testing')

__all__ = (archive.__all__ + comparators.__all__ + depositors.__all__ + exceptions.__all__ + historian.__all__ +
           process.__all__ + types.__all__ + helpers.__all__ + version.__all__ + history.__all__ +
           archive_factory.__all__ + refs.__all__ + records.__all__ + base_savable.__all__ + _ADDITIONAL)



from concurrent.futures import ThreadPoolExecutor, Future
from functools import partial

from PySide2 import QtCore, QtWidgets
from PySide2.QtCore import Signal

import mincepy
from . import common
from . import models
from . import tree_models

__all__ = 'TypeDropDown', 'ConnectionWidget', 'MincepyWidget'


class TypeDropDown(QtWidgets.QComboBox):
    """Drop down combo box that lists the types available in archive"""
    ALL = None

    # Signals
    selected_type_changed = Signal(object)

    def __init__(self, query_model: models.DataRecordQueryModel, parent=None):
        super().__init__(parent)
        self._query_model = query_model
        query_model.db_model.historian_changed.connect(self._update)

        self.setEditable(True)
        self._types = [None]
        self.addItem(self.ALL)

        def selection_changed(index):
            restrict_type = self._types[index]
            self.selected_type_changed.emit(restrict_type)

        self.currentIndexChanged.connect(selection_changed)

    @property
    def _historian(self):
        return self._query_model.db_model.historian

    def _update(self):
        self.clear()

        results = self._historian.get_archive().find()
        self._types = [None]
        self._types.extend(list(set(result.type_id for result in results)))

        type_names = self._get_type_names(self._types)

        self.addItems(type_names)
        completer = QtWidgets.QCompleter(type_names)
        self.setCompleter(completer)

    def _get_type_names(self, types):
        type_names = []
        for type_id in types:
            try:
                helper = self._historian.get_helper(type_id)
            except TypeError:
                type_names.append(str(type_id))
            else:
                type_names.append(mincepy.analysis.get_type_name(helper.TYPE))

        return type_names


class FilterControlPanel(QtWidgets.QWidget):
    # Signals
    display_as_class_changed = Signal(object)

    def __init__(self, entries_table: models.EntriesTable, parent=None):
        super().__init__(parent)
        self._entries_table = entries_table

        refresh = QtWidgets.QPushButton("Refresh")
        refresh.clicked.connect(self._entries_table.refresh)

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._create_type_drop_down())
        layout.addWidget(self._create_display_as_class_checkbox())
        layout.addWidget(refresh)
        self.setLayout(layout)

    def _create_display_as_class_checkbox(self):
        # Show snapshot class checkbox
        display_class_checkbox = QtWidgets.QCheckBox('Display as class', self)
        display_class_checkbox.setCheckState(QtCore.Qt.Checked)
        display_class_checkbox.stateChanged.connect(
            lambda state: self._entries_table.set_show_as_objects(state == QtCore.Qt.Checked))
        self._entries_table.set_show_as_objects(display_class_checkbox.checkState() == QtCore.Qt.Checked)

        return display_class_checkbox

    def _create_type_drop_down(self):
        type_drop_down = TypeDropDown(self._entries_table.query_model, self)
        type_drop_down.selected_type_changed.connect(self._entries_table.query_model.set_type_restriction)

        # Create an lay out the panel
        panel = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(QtWidgets.QLabel("Restrict type:"))
        layout.addWidget(type_drop_down)
        panel.setLayout(layout)

        return panel


class ConnectionWidget(QtWidgets.QWidget):
    # Signals
    connection_requested = Signal(str)
    historian_created = Signal(mincepy.Historian)

    def __init__(self,
                 default_connect_uri='',
                 create_historian_callback=common.default_create_historian,
                 executor=common.default_executor,
                 parent=None):
        super().__init__(parent)
        self._executor = executor
        self._create_historian = create_historian_callback

        self._connection_string = QtWidgets.QLineEdit(self)
        self._connection_string.setText(default_connect_uri)
        self._connect_button = QtWidgets.QPushButton('Connect', self)

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._connection_string)
        layout.addWidget(self._connect_button)
        self.setLayout(layout)

        self._connect_button.clicked.connect(self._connect_pushed)

    def _connect_pushed(self):
        uri = self._connection_string.text()
        self._executor(partial(self._connect, uri, ), "Connecting", blocking=True)

    def _connect(self, uri):
        try:
            historian = self._create_historian(uri)
        except Exception as exc:
            err_msg = "Error creating historian with uri '{}':\n{}".format(uri, exc)
            raise RuntimeError(err_msg)
        else:
            self.historian_created.emit(historian)

        return "Connected to {}".format(uri)


class MincepyWidget(QtWidgets.QWidget):

    def __init__(self, default_connect_uri='', create_historian_callback=common.default_create_historian,
                 executor=common.default_executor):
        super().__init__()

        self._create_historian_callback = create_historian_callback

        # The model
        self._db_model = models.DbModel()
        self._data_records = models.DataRecordQueryModel(self._db_model, executor=executor, parent=self)

        # Set up the connect panel of the GUI
        connect_panel = ConnectionWidget(default_connect_uri,
                                         create_historian_callback=create_historian_callback,
                                         executor=executor,
                                         parent=self)

        connect_panel.historian_created.connect(self._historian_created)

        self._entries_table = models.EntriesTable(self._data_records, parent=self)

        control_panel = FilterControlPanel(self._entries_table, self)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(connect_panel)
        self.layout.addWidget(control_panel)
        self.layout.addWidget(self._create_display_panel(self._entries_table))
        self.setLayout(self.layout)

    @property
    def db_model(self):
        return self._db_model

    def _historian_created(self, historian):
        self._db_model.historian = historian

    def _create_display_panel(self, entries_table: models.EntriesTable):
        panel = QtWidgets.QSplitter(QtCore.Qt.Vertical)

        entries_view = QtWidgets.QTableView(panel)
        entries_view.setSortingEnabled(True)
        entries_view.setModel(entries_table)

        record_tree = tree_models.RecordTree(parent=panel)
        record_tree_view = QtWidgets.QTreeView(panel)
        record_tree_view.setModel(record_tree)

        def row_changed(current, _previous):
            entries_table = self._entries_table
            record = entries_table.get_record(current.row())
            snapshot = entries_table.get_snapshot(current.row())
            record_tree.set_record(record, snapshot)

        entries_view.selectionModel().currentRowChanged.connect(row_changed)

        panel.addWidget(entries_view)
        panel.addWidget(record_tree_view)

        return panel


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, default_connect_uri='', create_historian_callback=None):
        super().__init__()
        self._executor = ThreadPoolExecutor()
        self._tasks = []

        self._main_widget = MincepyWidget(default_connect_uri, create_historian_callback, self._execute)
        self.setCentralWidget(self._main_widget)

        self._create_status_bar()

        self._task_done_signal.connect(self._task_done)

    def _create_status_bar(self):
        self.statusBar().showMessage('Ready')

    def _execute(self, func, msg=None, blocking=False) -> Future:
        future = self._executor.submit(func)
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor if blocking else QtCore.Qt.BusyCursor)
        self._tasks.append(future)
        future.add_done_callback(self._task_done_signal.emit)
        if msg is not None:
            self.statusBar().showMessage(msg)

        return future

    _task_done_signal = Signal(Future)

    @QtCore.Slot(Future)
    def _task_done(self, future):
        self._tasks.remove(future)
        QtWidgets.QApplication.restoreOverrideCursor()

        if not self._tasks:
            self.statusBar().clearMessage()

        try:
            new_msg = future.result()
            if new_msg is not None:
                self.statusBar().showMessage(new_msg, 1000)
        except Exception as exc:
            QtWidgets.QErrorMessage(self).showMessage(str(exc))



from abc import ABCMeta, abstractmethod
import collections
from typing import BinaryIO, Optional
import uuid

from . import base_savable
from . import depositors
from . import refs

__all__ = 'List', 'Str', 'Dict', 'BaseFile'


class _UserType(base_savable.BaseSavableObject):
    """Mixin for helping user types to be compatible with the historian.
    These typically have a .data member that stores the actual data (list, dict, str, etc)"""
    ATTRS = ('data',)
    data = None  # placeholder


class List(collections.UserList, _UserType):
    TYPE_ID = uuid.UUID('2b033f70-168f-4412-99ea-d1f131e3a25a')

    def save_instance_state(self, _saver: depositors.Saver):
        return self.data

    def load_instance_state(self, state, _loader: depositors.Loader):
        self.__init__(state)


class Dict(collections.UserDict, _UserType):
    TYPE_ID = uuid.UUID('a7584078-95b6-4e00-bb8a-b077852ca510')

    def save_instance_state(self, _saver: depositors.Saver):
        return self.data

    def load_instance_state(self, state, _loader: depositors.Loader):
        self.__init__(state)


class Str(collections.UserString, _UserType):
    TYPE_ID = uuid.UUID('350f3634-4a6f-4d35-b229-71238ce9727d')

    def save_instance_state(self, _saver: depositors.Saver):
        return self.data

    def load_instance_state(self, state, _loader: depositors.Loader):
        self.data = state


class RefList(collections.abc.MutableSequence, base_savable.BaseSavableObject):
    """A list that stores all entries as references in the database"""
    TYPE_ID = uuid.UUID('091efff5-136d-4ac2-bd59-28f50f151263')
    ATTRS = ('_data',)

    def __init__(self, init_list=None):
        super().__init__()
        self._data = []
        if init_list is not None:
            self._data = [refs.ObjRef(item) for item in init_list]

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)

    def __getitem__(self, item):
        return self._data[item]()

    def __setitem__(self, key, value):
        self._data[key] = refs.ObjRef(value)

    def __delitem__(self, key):
        self._data.__delitem__(key)

    def __len__(self):
        return self._data.__len__()

    def insert(self, index, value):
        self._data.insert(index, refs.ObjRef(value))


class RefDict(collections.MutableMapping, base_savable.BaseSavableObject):
    """A dictionary that stores all values as references in the database"""
    TYPE_ID = uuid.UUID('c95f4c4e-766b-4dda-a43c-5fca4fd7bdd0')
    ATTRS = ('_data',)

    def __init__(self, *args, **kwargs):
        super().__init__()
        initial = dict(*args, **kwargs)
        if initial:
            self._data = {key: refs.ObjRef(value) for key, value in initial.items()}
        else:
            self._data = {}

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)

    def __getitem__(self, item):
        return self._data[item]()

    def __setitem__(self, key, value):
        self._data[key] = refs.ObjRef(value)

    def __delitem__(self, key):
        self._data.__delitem__(key)

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()


class BaseFile(base_savable.BaseSavableObject, metaclass=ABCMeta):
    ATTRS = ('_filename', '_encoding')
    READ_SIZE = 256  # The number of bytes to read at a time

    def __init__(self, filename: str = None, encoding=None):
        super().__init__()
        self._filename = filename
        self._encoding = encoding

    @property
    def filename(self) -> Optional[str]:
        return self._filename

    @property
    def encoding(self) -> Optional[str]:
        return self._encoding

    @abstractmethod
    def open(self, mode='r', **kwargs) -> BinaryIO:
        """Open returning a file like object that supports close() and read()"""

    def __str__(self):
        contents = [str(self._filename)]
        if self._encoding is not None:
            contents.append("({})".format(self._encoding))
        return " ".join(contents)

    def __eq__(self, other) -> bool:
        """Compare the contents of two files

        If both files do not exist they are considered equal.
        """
        if not isinstance(other, BaseFile) or self.filename != other.filename:
            return False

        try:
            with self.open() as my_file:
                try:
                    with other.open() as other_file:
                        while True:
                            my_line = my_file.readline(self.READ_SIZE)
                            other_line = other_file.readline(self.READ_SIZE)
                            if my_line != other_line:
                                return False
                            if my_line == '' and other_line == '':
                                return True
                except FileNotFoundError:
                    return False
        except FileNotFoundError:
            # Our file doesn't exist, make sure the other doesn't either
            try:
                with other.open():
                    return False
            except FileNotFoundError:
                return True

    def yield_hashables(self, hasher):
        """Has the contents of the file"""
        try:
            with self.open('rb') as opened:
                while True:
                    line = opened.read(self.READ_SIZE)
                    if line == b'':
                        return
                    yield line
        except FileNotFoundError:
            yield from hasher.yield_hashables(None)


HISTORIAN_TYPES = Str, List, Dict, RefList, RefDict



import sys

from PySide2 import QtWidgets

import mincepy
from . import views

__all__ = ('run_application',)


def run_application(default_connect_uri="mongodb://localhost/test_database",
                    create_historian_callback=mincepy.create_historian):
    app = QtWidgets.QApplication([])
    widget = views.MainWindow(default_connect_uri, create_historian_callback)
    widget.resize(800, 600)
    widget.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    run_application()



"""References module"""
import uuid

from . import records
from . import types

__all__ = 'ObjRef', 'auto_deref'


def auto_deref(obj):
    if isinstance(obj, ObjRef):
        return obj()
    if isinstance(obj, dict):
        return {key: auto_deref(value) for key, value in obj.items()}
    if isinstance(obj, list):
        return [auto_deref(item) for item in obj]
    return obj


class ObjRef(types.SavableObject):
    TYPE_ID = uuid.UUID('633c7035-64fe-4d87-a91e-3b7abd8a6a28')

    def __init__(self, obj, auto=False):
        super().__init__()
        self._obj = obj
        self._auto = auto
        self._ref = None
        self._loader = None

    @property
    def auto(self):
        return self._auto

    def __str__(self):
        desc = "ref: "
        if self._obj is not None:
            desc += str(self._obj)
        else:
            desc += str(self._ref)
        return desc

    def __call__(self):
        if self._obj is None:
            if self._ref is None:
                raise RuntimeError("Cannot dereference a None reference")
            # Cache the object
            self._obj = self._loader.load(self._ref)
            assert self._obj is not None, "Loader did not load object"
            self._ref = None
            self._loader = None

        return self._obj

    def __eq__(self, other):
        return self is other

    def yield_hashables(self, hasher):
        yield from hasher.yield_hashables(self._ref)

    def save_instance_state(self, saver):
        ref = self._ref
        if ref is None:
            # This should mean that we have loaded the object via the 'call' method previously
            assert self._obj is not None
            ref = saver.ref(self._obj)

        return {'ref': [ref.obj_id, ref.version], 'auto': self._auto}

    def load_instance_state(self, saved_state, loader):
        self._ref = records.Ref(*saved_state['ref'])
        self._obj = None
        self._loader = loader
        self._auto = saved_state['auto']


HISTORIAN_TYPES = (ObjRef,)



from argparse import Namespace

import mincepy
import mincepy.records
from mincepy.testing import Car, Cycle


def test_obj_ref_simple(historian: mincepy.Historian):
    a = Cycle()
    a.ref = mincepy.ObjRef(a)
    aid = historian.save(a)
    del a

    loaded = historian.load(aid)
    assert loaded.ref() is loaded


def test_obj_ref_snapshot(historian: mincepy.Historian):
    """Check that a historic snapshot still works with references"""
    ns = Namespace()
    car = Car('honda', 'white')
    ns.car = mincepy.ObjRef(car)
    ns_id_honda = historian.save(ns, return_sref=True)

    car.make = 'fiat'
    ns_id_fiat = historian.save(ns, return_sref=True)
    del ns

    assert ns_id_fiat.version == ns_id_honda.version + 1

    loaded = historian.load(ns_id_honda)
    assert loaded.car().make == 'honda'

    loaded2 = historian.load(ns_id_fiat)
    assert loaded2.car().make == 'fiat'
    assert loaded2.car() is not car

    # Load the 'live' namespace
    loaded3 = historian.load(ns_id_honda.obj_id)
    assert loaded3.car() is car


def test_obj_ref_complex(historian: mincepy.Historian):
    honda = Car('honda')
    nested1 = Namespace()
    nested2 = Namespace()
    parent = Namespace()

    # Both the nested refer to the same car
    nested1.car = mincepy.ObjRef(honda)
    nested2.car = mincepy.ObjRef(honda)

    # Now put them in their containers
    parent.ns1 = mincepy.ObjRef(nested1)
    parent.ns2 = mincepy.ObjRef(nested2)

    parent_id = historian.save(parent)
    del parent

    loaded = historian.load(parent_id)
    assert loaded.ns1() is nested1
    assert loaded.ns2() is nested2
    assert loaded.ns1().car() is loaded.ns2().car()

    fiat = Car('fiat')
    loaded.ns2().car = mincepy.ObjRef(fiat)
    parent_snapshot_id = historian.save(loaded, return_sref=True)
    del loaded

    loaded2 = historian.load_snapshot(mincepy.records.Ref(parent_id, 0))
    assert loaded2.ns1().car().make == 'honda'
    assert loaded2.ns2().car().make == 'honda'
    del loaded2

    loaded3 = historian.load_snapshot(parent_snapshot_id)
    assert loaded3.ns1().car().make == 'honda'
    assert loaded3.ns2().car().make == 'fiat'



from typing import Optional

from . import builtins
from . import common_helpers
from . import historian
from . import refs

__all__ = 'get_historian', 'set_historian'

CURRENT_HISTORIAN = None


def get_historian() -> historian.Historian:
    global CURRENT_HISTORIAN  # pylint: disable=global-statement
    return CURRENT_HISTORIAN


def set_historian(new_historian: Optional[historian.Historian], register_common_types=True):
    global CURRENT_HISTORIAN  # pylint: disable=global-statement
    CURRENT_HISTORIAN = new_historian
    if new_historian is not None and register_common_types:
        new_historian.register_types(common_helpers.HISTORIAN_TYPES)
        new_historian.register_types(builtins.HISTORIAN_TYPES)
        new_historian.register_types(refs.HISTORIAN_TYPES)
